/* 
 * $Id: bgpsim.c,v 1.13 1998/07/23 17:57:19 labovit Exp $
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>

#include <mrt.h>
#include <timer.h>
#include <io.h>
#include <interface.h>
#include <bgp.h>
#include <rip.h>
#ifdef HAVE_IPV6
#include <ripng.h>
#endif /* HAVE_IPV6 */
#include <bgp.h>
#include <config_file.h>
#include "config.h"
#include "bgpsim.h"


simulation_t *SIMULATION;
bgp_t *BGP;
io_t *IO;
extern int BGPSIM;
extern int BGPSIM_TRANSPARENT;
gateway_t *local_gateway;
trace_t *default_trace;
#ifdef HAVE_IPV6
gateway_t *local_gateway6;
#endif /* HAVE_IPV6 */

/* don't support yet, but linking libraries may require */
rip_t *RIP;
#ifdef HAVE_IPV6
ripng_t *RIPNG;
#endif /* HAVE_IPV6 */

int show_simulation (uii_connection_t * uii);
void stop_simulation_schedule (uii_connection_t * uii);


/* main
 */
void 
main (int argc, char *argv[])
{
    char c;
    extern char *optarg;	/* getopt stuff */
    extern int optind;		/* getopt stuff */
    int errors = 0;
    u_long id = htonl (DEFAULT_MY_ID);

    char *usage = "Usage: bgpsim [-i router_id] [-f file] [-v]\n";
    char *config_file = "./bgpsim.conf";
    default_trace = New_Trace ();

    /* must be root -- I guess the BGP code requires this somewhere?? */
    if (getuid ()) {
      fprintf (stderr, "bgpsim: must be root\n");
      exit (1);
    }

    /* init simulation */
    SIMULATION = New (simulation_t);
    SIMULATION->ll_networks = LL_Create (0);
    BGPSIM_TRANSPARENT = 1;

    while ((c = getopt (argc, argv, "vf:i:s")) != -1)
      switch (c) {
	case 'v':		/* verbose */
	    set_trace (default_trace, TRACE_FLAGS, TR_ALL,
		       TRACE_LOGFILE, "stdout",
		       NULL);
	    break;
	case 'f':		/* config file */
	    config_file = strdup (optarg);
	    break;
	case 'i':
	    if (inet_pton (AF_INET, optarg, &id) > 0)
	        break;
        case 's':  /* by default, bgpsim is transparent. set nexthop, origin... */
	  BGPSIM_TRANSPARENT = 0;
	  break;
	case 'h':
	default:
	    errors++;
	    break;
	}

    if (errors) {
      fprintf (stderr, usage);
	printf ("\nBGPSIM version (%s) compiled on %s\n\n",
		MRT_VERSION, __DATE__);
	exit (0);
    }

    BGPSIM = 1;

    init_mrt (default_trace);
    init_mrt_reboot (argc, argv);

    init_interfaces (default_trace); 
    init_mrtd_config (default_trace);
    add_bgpsim_config (default_trace);

    init_BGP (default_trace);
    set_BGP (BGP_MY_ID, id, 0);

    if (config_from_file (default_trace, config_file) < 0) {
      UII->initial_state = 1;   /* no password */
      config_create_default ();
    }

    local_gateway = add_gateway (ascii2prefix (AF_INET, "127.0.0.1"), 
				 BGP->my_as, NULL);
#ifdef HAVE_IPV6
    local_gateway6 = add_gateway (ascii2prefix (AF_INET6, "::1"), 
				  BGP->my_as, NULL);
#endif /* HAVE_IPV6 */

    set_uii (UII, UII_PROMPT, UII_NORMAL, "Bgpsim> ", 0);
    /* set_uii (UII, UII_PROMPT, UII_CONFIG, "Bgpsim-Config> ", 0); */

    /*   signal (SIGINT, sbgp_quit); */

    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "show simulation", 
		     show_simulation, "Show status of simulation");
    uii_add_command2 (UII_NORMAL, COMMAND_NORM, "stop", 
		     stop_simulation_schedule, "Stop simulation");

    running = 1;
    listen_uii2 ("mrtd");

    start_bgp ();
    start_simulation (default_trace);
    mrt_main_loop ();
    /* NOT REACHED */
}


